use std::cmp::{max, min};
use crate::position::Position;

pub const LEVEL_WIDTH: usize = 50;
pub const LEVEL_HEIGHT: usize = 25;

#[derive(Copy, Clone, Debug, PartialEq)]
pub enum LevelElement {
    Wall,
    Floor,
    StairDown,
    StairUp,
    Unknown,
}

#[derive(Copy, Clone, Debug, PartialEq)]
pub enum RoomType {
    Start,
    End,
    StairUp,
    StairDown,
    BasicRoom,
    EmptyRoom,
}

#[derive(Copy, Clone, Debug, PartialEq)]
pub struct Level {
    pub(crate) structure: [[LevelElement; LEVEL_HEIGHT]; LEVEL_WIDTH],
    pub(crate) discovered: [[bool; LEVEL_HEIGHT]; LEVEL_WIDTH],
    pub(crate) start_x: usize,
    pub(crate) start_y: usize,
    pub(crate) end_x: usize,
    pub(crate) end_y: usize,
}

impl Level {
    pub fn new() -> Level {
        let s = [[LevelElement::Wall; LEVEL_HEIGHT]; LEVEL_WIDTH];
        Level {
            structure: s,
            discovered: [[false; LEVEL_HEIGHT]; LEVEL_WIDTH],
            start_x: 1,
            start_y: 1,
            end_x: 1,
            end_y: 1,
        }
    }
    pub fn get_element(&self, x: i16, y: i16) -> Option<LevelElement> {
        if x < 0 || y < 0 {
            return None;
        }
        let x = x as usize;
        let y = y as usize;
        if x >= LEVEL_WIDTH || y >= LEVEL_HEIGHT {
            return None;
        }
        if self.discovered[x][y] {
            return Some(self.structure[x][y]);
        }
        return Some(LevelElement::Unknown);
    }
    /// discover the area with in the level around the given position
    pub fn discover(&mut self, pos: Position) {
        let x = pos.get_x();
        let y = pos.get_y();
        for x_r in max(x - 1, 0)..=min(x + 1, LEVEL_WIDTH - 1) {
            for y_r in max(y - 1, 0)..=min(y + 1, LEVEL_HEIGHT - 1) {
                self.discovered[x_r][y_r] = true;
            }
        }
        for x_r in x..=min(x + 2, LEVEL_WIDTH - 1) {
            self.discovered[x_r][y] = true;
            if self.structure[x_r][y] == LevelElement::Wall {
                break;
            }
        }
        for y_r in y..=min(y + 2, LEVEL_HEIGHT - 1) {
            self.discovered[x][y_r] = true;
            if self.structure[x][y_r] == LevelElement::Wall {
                break;
            }
        }
        for x_r in (max(x, 2)..x + 2).rev() {
            self.discovered[x_r - 2][y] = true;
            if self.structure[x_r - 2][y] == LevelElement::Wall {
                break;
            }
        }
        for y_r in (max(y, 2)..y + 2).rev() {
            self.discovered[x][y_r - 2] = true;
            if self.structure[x][y_r - 2] == LevelElement::Wall {
                break;
            }
        }
    }
}

#[test]
fn test_get_element() {
    let l = Level::new();
    assert_eq!(l.get_element(0, 0).unwrap(), LevelElement::Wall);
    assert_eq!(l.get_element(4, 4).unwrap(), LevelElement::StairDown);
    assert_eq!(l.get_element(49, 24).unwrap(), LevelElement::Wall);
    assert_eq!(l.get_element(50, 25), None);
}
