use ratatui::prelude::*;
use ratatui::widgets::Widget;

use crate::game::Game;
use crate::level::{Level, LevelElement};

const FG_BROWN: Color = Color::Rgb(186, 74, 0);

pub struct LevelWidget<'a> {
    pub game: &'a Game,
}

impl<'a> LevelWidget<'a> {
    fn set_cell(&self, buf: &mut Buffer, x: u16, y: u16, symbol: &str, fg: Color, bg: Color) {
        buf.
            get_mut(x, y).
            set_symbol(symbol).
            set_bg(bg).
            set_fg(fg);
    }

    fn calculate_wall(top: Option<LevelElement>, right: Option<LevelElement>, bottom: Option<LevelElement>, left: Option<LevelElement>) -> String {
        let mut val = 0;
        val += match top {
            Some(n) => {
                match n {
                    LevelElement::Wall => 1,
                    _ => 0
                }
            }
            None => 0
        };
        val += match right {
            Some(n) => {
                match n {
                    LevelElement::Wall => 2,
                    _ => 0
                }
            }
            None => 0
        };
        val += match bottom {
            Some(n) => {
                match n {
                    LevelElement::Wall => 4,
                    _ => 0
                }
            }
            None => 0
        };
        val += match left {
            Some(n) => {
                match n {
                    LevelElement::Wall => 8,
                    _ => 0
                }
            }
            None => 0
        };
        [
            "╋", "┃", "━", "┗",
            "┃", "┃", "┏", "┣",
            "━", "┛", "━", "┻",
            "┓", "┫", "┳", "╋",
        ][val].to_string()
    }
}

impl<'a> Widget for LevelWidget<'a> {
    fn render(self, area: Rect, buf: &mut Buffer) {
        let player = &self.game.player;
        let player_pos = player.get_position();
        let level: &Level = &self.game.levels[player_pos.get_level()];

        let al: u16 = area.left();
        let at: u16 = area.top();

        // draw the level elements
        for x in al..area.right() {
            for y in at..area.bottom() {
                let level_x = (x - al) as i16;
                let level_y = (y - at) as i16;
                match level.get_element(level_x, level_y) {
                    None => {}
                    Some(e) => {
                        match e {
                            LevelElement::Wall => {
                                let top = level.get_element(level_x, level_y - 1);
                                let right = level.get_element(level_x + 1, level_y);
                                let bottom = level.get_element(level_x, level_y + 1);
                                let left = level.get_element(level_x - 1, level_y);
                                let symbol = LevelWidget::calculate_wall(top, right, bottom, left);
                                self.set_cell(buf, x, y, symbol.as_str(), FG_BROWN, Color::Gray);
                            }
                            LevelElement::Floor => {
                                self.set_cell(buf, x, y, " ", FG_BROWN, Color::Gray);
                            }
                            LevelElement::StairDown => {
                                self.set_cell(buf, x, y, ">", Color::Black, Color::Gray);
                            }
                            LevelElement::StairUp => {
                                self.set_cell(buf, x, y, "<", Color::Black, Color::Gray);
                            }
                            LevelElement::Unknown => {
                                self.set_cell(buf, x, y, "▒", Color::DarkGray, Color::Gray);
                            }
                        }
                    }
                }
            }
        }

        // draw player
        let player_x = al + player_pos.get_x() as u16;
        let player_y = at + player_pos.get_y() as u16;
        self.set_cell(buf, player_x, player_y, "8", Color::Red, Color::Gray);
    }
}

#[test]
fn test_get_wall_symbol() {
    let wall = Some(LevelElement::Wall);
    // 0
    let s = LevelWidget::calculate_wall(None, None, None, None);
    assert_eq!(s, "╋");
    // 15
    let s = LevelWidget::calculate_wall(wall, wall, wall, wall);
    assert_eq!(s, "╋");

    // 1
    let s = LevelWidget::calculate_wall(wall, None, None, None);
    assert_eq!(s, "┃");
    // 5
    let s = LevelWidget::calculate_wall(wall, None, wall, None);
    assert_eq!(s, "┃");
    // 4
    let s = LevelWidget::calculate_wall(None, None, wall, None);
    assert_eq!(s, "┃");

    // 2
    let s = LevelWidget::calculate_wall(None, wall, None, None);
    assert_eq!(s, "━");
    // 10
    let s = LevelWidget::calculate_wall(None, wall, None, wall);
    assert_eq!(s, "━");

    // 9
    let s = LevelWidget::calculate_wall(wall, None, None, wall);
    assert_eq!(s, "┛");
    // 3
    let s = LevelWidget::calculate_wall(wall, wall, None, None);
    assert_eq!(s, "┗");
    // 6
    let s = LevelWidget::calculate_wall(None, wall, wall, None);
    assert_eq!(s, "┏");
    // 12
    let s = LevelWidget::calculate_wall(None, None, wall, wall);
    assert_eq!(s, "┓");
}

